==> blcr-0.2.0.tar.gz: blcr-0.2.0/cr_module/cr_dest_file.c <==
/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_dest_file.c,v 1.14 2003/11/14 19:55:14 phargrov Exp $
 */

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/slab.h>

#include "cr_module.h"

#define CR_FILE_PATT	"context.%d"
#define CR_FILE_MAX	32	// XXX: better size?
#define CR_FILE_MODE	0600


// XXX: need comment here
// Note: we fput if appropriate.
//
// XXX: Are these checks sufficient/correct?
//
// XXX:  These need to return good error codes!
static int validate_file(struct file *file)
{
	int retval = -EINVAL;
	struct dentry *dentry;
	struct inode *inode;

	if (file == NULL) {
	        retval = -EBADF;
		// Fall through to failure
	} else if (IS_ERR(file)) {
		retval = PTR_ERR(file);
		// Fall through to failure
	} else {
		dentry = file->f_dentry;
		inode = dentry->d_inode;
	
		if (inode->i_nlink > 1) {
		        retval = -EMLINK;
			// Don't dump to a file w/ multiple links
			// Fall through to failure
		} else if (d_unhashed(dentry)) {
			// Fail due to NFS or autofs timeouts.
			// Fall through to failure
		} else if (!S_ISREG(inode->i_mode) &&
			   !S_ISSOCK(inode->i_mode) &&
			   !S_ISFIFO(inode->i_mode)) {
			// Fall through to failure
		} else if (!file->f_op || !file->f_op->write) {
		        retval = -EROFS;
			// Fail because there is no write method
			// Fall through to failure
		} else if(!(file->f_mode & FMODE_WRITE)) {
		        retval = -EACCES;
			// Fail because there is no write permission
			// Fall through to failure
		} else {
			// SUCCESS:
			return 0;
		}
	
		// Only failure exits here
		fput(file);
	}

	// Only failure exits here
	return retval;
}

// build_fs_struct()
//
// Build an fs_struct which will later allow us to create a file in a
// specified directory.
//
// Based on kernel/fork.c:__copy_fs_struct()
//
static inline struct fs_struct *build_fs_struct(struct fs_struct *old,
						struct file *dirp)
{
	struct fs_struct *fs = kmalloc(sizeof(*fs), GFP_KERNEL);
	if (fs) {
		// Copy MOST of 'old', but not the pwd
		atomic_set(&fs->count, 1);
		fs->lock = RW_LOCK_UNLOCKED;
		fs->umask = old->umask;
		read_lock(&old->lock);
		fs->rootmnt = mntget(old->rootmnt);
		fs->root = dget(old->root);
		if (old->altroot) {
			fs->altrootmnt = mntget(old->altrootmnt);
			fs->altroot = dget(old->altroot);
		} else {
			fs->altrootmnt = NULL;
			fs->altroot = NULL;
		}	
		read_unlock(&old->lock);

		// Get pwd from 'dirp' rather than from 'old'
		fs->pwdmnt = mntget(dirp->f_vfsmnt);
		fs->pwd = dget(dirp->f_dentry);
	}
	return fs;
}

// free_fs_struct()
//
// Release resources allocated by build_fs_struct().
//
static inline void free_fs_struct(struct fs_struct *fs)
{
	dput(fs->root);
	mntput(fs->rootmnt);
	dput(fs->pwd);
	mntput(fs->pwdmnt);
	if (fs->altroot) {
		dput(fs->altroot);
		mntget(fs->altrootmnt);
	}
	kfree(fs);
}

// XXX: need comment here
static inline int dest_init_reg(cr_chkpt_dest_t *dest, struct file *filp)
{
	int result;

	result = validate_file(filp);
	if (!result) {
	    dest->filp = filp;
	    init_MUTEX(&dest->mutex);
	}

	return result;
}

// XXX: need comment here
// XXX: currently only check perms to create here.
// XXX: other checks needed?
static inline int dest_init_dir(cr_chkpt_dest_t *dest, struct file *dirp)
{
	int result = 0;

	result = permission(dirp->f_dentry->d_inode, MAY_WRITE | MAY_EXEC);
	if (!result) {
		dest->fs = build_fs_struct(current->fs, dirp);
	}

	// Error and normal paths exit here
	fput(dirp);	// We don't hold the filp for a directory
	return result;
}

// cr_dest_init(dest, fd, from)
//
// Validate and record data about the requested destination of a checkpoint
//
// Returns 0 on success, negative error code on failure.
//
// XXX: need to document return cases?
// XXX: need to be sure we make all the right checks.
int cr_dest_init(cr_chkpt_dest_t *dest, int fd, struct file *from)
{
	struct file *filp;

	memset(dest, 0, sizeof(*dest));

	if (fd != CR_DEST_CWD) {
		filp = fget(fd);
	} else {
		// XXX: do we want to shortcut and just copy current->fs?
		filp = filp_open(".", O_RDONLY|O_NDELAY|O_DIRECTORY, 0);
	}

	if (!filp) {
	        CR_KTRACE_BADPARM("invalid file handle %d received", fd);
		return -EINVAL;
	} else if (IS_ERR(filp)) {
		return PTR_ERR(filp);
	} else if (filp == from) {
	        CR_KTRACE_BADPARM("filp == from fd! %d", fd);
		fput(filp);
		return -EINVAL;
	}

	switch (filp->f_dentry->d_inode->i_mode & S_IFMT) {
	case S_IFREG: case S_IFIFO: case S_IFSOCK:
	        CR_KTRACE_LOW_LVL("Calling dest_init_reg on fd %d", fd);
		return dest_init_reg(dest, filp);
		break;

	case S_IFDIR:
		return dest_init_dir(dest, filp);
		break;

	case S_IFCHR: case S_IFBLK:
		// We don't deal with these cases yet:
		// fall through...

	default:
	        CR_KTRACE_BADPARM("unsupported file type");
		fput(filp);
		return -EINVAL;
	}
}

// cr_dest_free(dest)
//
// Free the checkpoint destination, decrementing use counts as appropriate.
void cr_dest_free(cr_chkpt_dest_t *dest)
{
	// For file/socket, etc.:
	if (dest->filp) {
		fput(dest->filp);
	}

	// For directory:
	if (dest->fs) {
		free_fs_struct(dest->fs);
	}
}

// cr_dest_get(dest, shared)
//
// Returns the filp to write to.
// Sets 'shared' to indicate if the file is shared with other threads.
// If so, then the caller is responsible for any synchronization.
// The dest->mutex field is provided for this purpose.
//
struct file *cr_dest_get(cr_chkpt_dest_t *dest, int *shared)
{
	struct file *filp = NULL;

	if (dest->filp) {
		// Destination is a "regular" file.

		// Just return the filp.  We don't need to fget()
		// because the fget() done in cr_dest_init is sufficient.
		filp = dest->filp;

		if (shared) {
			*shared = 1;
		}
	} else if (dest->fs) {
		// Destination is a per-process file in a given directory.

		struct fs_struct *saved_fs;
		char filename[CR_FILE_MAX];
		int error;

		// Create a filename
		snprintf(filename, CR_FILE_MAX,
			 CR_FILE_PATT, current->pid);

		// Play with current->fs to open() in the destination dir
		saved_fs = current->fs;
		current->fs = dest->fs;
		filp = filp_open(filename, // 2 = write only
				 2 | O_CREAT | O_NOFOLLOW | O_TRUNC,
				 CR_FILE_MODE);
		current->fs = saved_fs;

		error = validate_file(filp);
		if (error) {
			filp = ERR_PTR(error);
		}

		if (shared) {
			*shared = 0;
		}
	}

	return filp;
}

// cr_dest_put()
//
// Put the filp written to.
// This is always paired with cr_dest_get() calls.
//
void cr_dest_put(cr_chkpt_dest_t *dest, struct file *filp)
{
	if (dest->fs) {
		// Dest is a directory.  Release the temporary filp.
		fput(filp);
	} if (dest->filp) {
		// Dest is a "regular" file.  DO NOTHING
	}
}

==> blcr-0.6.2.tar.gz: blcr-0.6.2/cr_module/cr_dest_file.c <==
/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_dest_file.c,v 1.19 2007/01/05 23:43:21 phargrov Exp $
 */

#include "cr_module.h"

#define CR_FILE_PATT	"context.%d"
#define CR_FILE_MAX	32	// XXX: better size?
#define CR_FILE_MODE	0600


// XXX: need comment here
// Note: we fput if appropriate.
//
// XXX: Are these checks sufficient/correct?
//
// XXX:  These need to return good error codes!
static int validate_file(struct file *file)
{
	int retval = -EINVAL;
	struct dentry *dentry;
	struct inode *inode;

	if (file == NULL) {
	        retval = -EBADF;
		// Fall through to failure
	} else if (IS_ERR(file)) {
		retval = PTR_ERR(file);
		// Fall through to failure
	} else {
		dentry = file->f_dentry;
		inode = dentry->d_inode;
	
		if (inode->i_nlink > 1) {
		        retval = -EMLINK;
			// Don't dump to a file w/ multiple links
			// Fall through to failure
		} else if (d_unhashed(dentry)) {
			// Fail due to NFS or autofs timeouts.
			// Fall through to failure
		} else if (!S_ISREG(inode->i_mode) &&
			   !S_ISCHR(inode->i_mode) &&
			   !S_ISSOCK(inode->i_mode) &&
			   !S_ISFIFO(inode->i_mode)) {
			// Fall through to failure
		} else if(!(file->f_mode & FMODE_WRITE)) {
		        retval = -EBADF;
			// Fail because there is no write permission
			// Fall through to failure
		} else {
			// SUCCESS:
			return 0;
		}
	
		// Only failure exits here
		fput(file);
	}

	// Only failure exits here
	return retval;
}

// build_fs_struct()
//
// Build an fs_struct which will later allow us to create a file in a
// specified directory.
//
// Based on kernel/fork.c:__copy_fs_struct()
//
static inline struct fs_struct *build_fs_struct(struct fs_struct *old,
						struct file *dirp)
{
	struct fs_struct *fs = kmalloc(sizeof(*fs), GFP_KERNEL);
	if (fs) {
		// Copy MOST of 'old', but not the pwd
		atomic_set(&fs->count, 1);
		fs->lock = RW_LOCK_UNLOCKED;
		fs->umask = old->umask;
		read_lock(&old->lock);
		fs->rootmnt = mntget(old->rootmnt);
		fs->root = dget(old->root);
		if (old->altroot) {
			fs->altrootmnt = mntget(old->altrootmnt);
			fs->altroot = dget(old->altroot);
		} else {
			fs->altrootmnt = NULL;
			fs->altroot = NULL;
		}	
		read_unlock(&old->lock);

		// Get pwd from 'dirp' rather than from 'old'
		fs->pwdmnt = mntget(dirp->f_vfsmnt);
		fs->pwd = dget(dirp->f_dentry);
	}
	return fs;
}

// free_fs_struct()
//
// Release resources allocated by build_fs_struct().
//
static inline void free_fs_struct(struct fs_struct *fs)
{
	dput(fs->root);
	mntput(fs->rootmnt);
	dput(fs->pwd);
	mntput(fs->pwdmnt);
	if (fs->altroot) {
		dput(fs->altroot);
		mntget(fs->altrootmnt);
	}
	kfree(fs);
}

// XXX: need comment here
static inline int dest_init_reg(cr_location_t *dest, struct file *filp)
{
	int result;

	result = validate_file(filp);
	if (!result) {
	    dest->filp = filp;
	    init_MUTEX(&dest->mutex);
	}

	return result;
}

// XXX: need comment here
// XXX: currently only check perms to create here.
// XXX: other checks needed?
static inline int dest_init_dir(cr_location_t *dest, struct file *dirp)
{
	int result = 0;

	result = cr_permission(dirp->f_dentry->d_inode, MAY_WRITE | MAY_EXEC, NULL);
	if (!result) {
		dest->fs = build_fs_struct(current->fs, dirp);
	}

	// Error and normal paths exit here
	fput(dirp);	// We don't hold the filp for a directory
	return result;
}

// cr_dest_init(dest, fd, from)
//
// Validate and record data about the requested destination of a checkpoint
//
// Returns 0 on success, negative error code on failure.
//
// XXX: need to document return cases?
// XXX: need to be sure we make all the right checks.
int cr_dest_init(cr_location_t *dest, int fd, struct file *from)
{
	struct file *filp;

	memset(dest, 0, sizeof(*dest));

	if (fd != CR_DEST_CWD) {
		filp = fget(fd);
	} else {
		// XXX: do we want to shortcut and just copy current->fs?
		filp = filp_open(".", O_RDONLY|O_NDELAY|O_DIRECTORY, 0);
	}

	if (!filp) {
	        CR_KTRACE_BADPARM("invalid file handle %d received", fd);
		return -EINVAL;
	} else if (IS_ERR(filp)) {
		return PTR_ERR(filp);
	} else if (filp == from) {
	        CR_KTRACE_BADPARM("filp == from fd! %d", fd);
		fput(filp);
		return -EINVAL;
	}

	switch (filp->f_dentry->d_inode->i_mode & S_IFMT) {
	case S_IFREG: case S_IFCHR: case S_IFIFO: case S_IFSOCK:
	        CR_KTRACE_LOW_LVL("Calling dest_init_reg on fd %d", fd);
		return dest_init_reg(dest, filp);
		break;

	case S_IFDIR:
		return dest_init_dir(dest, filp);
		break;

	case S_IFBLK:
		// We don't deal with this case yet:
		// fall through...

	default:
	        CR_KTRACE_BADPARM("unsupported file type");
		fput(filp);
		return -EINVAL;
	}
}

// cr_dest_free(dest)
//
// Free the checkpoint destination, decrementing use counts as appropriate.
void cr_dest_free(cr_location_t *dest)
{
	// For file/socket, etc.:
	if (dest->filp) {
		fput(dest->filp);
	}

	// For directory:
	if (dest->fs) {
		free_fs_struct(dest->fs);
	}
}

// cr_dest_get(dest, shared)
//
// Returns the filp to write to.
// Sets 'shared' to indicate if the file is shared with other threads.
// If so, then the caller is responsible for any synchronization.
// The dest->mutex field is provided for this purpose.
//
struct file *cr_dest_get(cr_location_t *dest, int *shared)
{
	struct file *filp = NULL;

	if (dest->filp) {
		// Destination is a "regular" file.

		// Just return the filp.  We don't need to fget()
		// because the fget() done in cr_dest_init is sufficient.
		filp = dest->filp;

		if (shared) {
			*shared = 1;
		}
	} else if (dest->fs) {
		// Destination is a per-process file in a given directory.

		struct fs_struct *saved_fs;
		char filename[CR_FILE_MAX];
		int error;

		// Create a filename
		sprintf(filename, CR_FILE_PATT, current->pid);

		// Play with current->fs to open() in the destination dir
		saved_fs = current->fs;
		current->fs = dest->fs;
		filp = filp_open(filename, // 2 = write only
				 2 | O_CREAT | O_NOFOLLOW | O_TRUNC,
				 CR_FILE_MODE);
		current->fs = saved_fs;

		error = validate_file(filp);
		if (error) {
			filp = ERR_PTR(error);
		}

		if (shared) {
			*shared = 0;
		}
	}

	return filp;
}

// cr_dest_put()
//
// Put the filp written to.
// This is always paired with cr_dest_get() calls.
//
void cr_dest_put(cr_location_t *dest, struct file *filp)
{
	if (dest->fs) {
		// Dest is a directory.  Release the temporary filp.
		fput(filp);
	} if (dest->filp) {
		// Dest is a "regular" file.  DO NOTHING
	}
}
